/*
 * Copyright (c) 2006, Adam Dunkels
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * ---------------------------------------------------------------------------
 * Source modified by Uwe Berger (bergeruw@gmx.net); 2010
 * Additional modification by Rene Bllhoff (reneboellhoff@gmx.de); Jun 2010
 * ---------------------------------------------------------------------------
 */

#define DEBUG 0

#if DEBUG
#define DEBUG_PRINTF(...)  printf(__VA_ARGS__)
#else
#define DEBUG_PRINTF(...)
#endif


#include "platform.h"

#include "ubasic.h"
#include "tokenizer.h"

#include <stdio.h> /* printf() */
#include <stdlib.h> /* exit() */



void FATAL_ABORT(void)
{
	printf ("Fatal Abort\n\r");
}


char const *program_ptr;
char string[MAX_STRINGLEN];

int gosub_stack[MAX_GOSUB_STACK_DEPTH];
int gosub_stack_ptr;

struct for_state {
  int line_after_for;
  int for_variable;
  int to;
};

struct for_state for_stack[MAX_FOR_STACK_DEPTH];
int for_stack_ptr;
char variables[MAX_VARNUM];

int ended;

int  expr(void);
void line_statement(void);
void statement(void);
int get_parameters(T_USER_FUNC_DATAS *params, int numparams, int next);
int get_arrayindex (void);
void accept(int token);

/*---------------------------------------------------------------------------*/
void ubasic_init(const char *program)
{
  program_ptr = program;
  for_stack_ptr = gosub_stack_ptr = 0;
  tokenizer_init(program);
  ended = 0;
}
/*---------------------------------------------------------------------------*/
int get_parameters(T_USER_FUNC_DATAS *params, int numparams, int next)
{
	unsigned char ptr = 0;
	if (numparams == 0) { tokenizer_next(); return 0;} 

	while (ptr < numparams)
	{
		switch (tokenizer_token())
		{
			case TOKENIZER_STRING 	:
				tokenizer_string (((*params) [ptr].text), MAX_USER_KEYWORD_LEN);
				(*params) [ptr].type = 2;
				tokenizer_next ();
				break;
			case TOKENIZER_NUMBER 	:
			case TOKENIZER_VARIABLE :
			case TOKENIZER_NAME 		:
				(*params) [ptr].value = expr();
				(*params) [ptr].type = 1;
				break;
			default :
				DEBUG_PRINTF("Unexpected end of parameter list\n");
				FATAL_ABORT();
		}
		if (++ptr < numparams) { accept (TOKENIZER_COMMA); }
	}
	if (next) { tokenizer_next(); }
	return 0;
}
/*---------------------------------------------------------------------------*/
int get_arrayindex (void)
{
	int i;
	accept (TOKENIZER_LEFTBRACK);
	i = expr();
	accept (TOKENIZER_RIGHTBRACK);
	return i;
}
/*---------------------------------------------------------------------------*/
void accept(int token)
{
  if(token != tokenizer_token()) {
    DEBUG_PRINTF("Token not what was expected (expected %d, got %d)\n",
		 token, tokenizer_token());
    tokenizer_error_print();
    FATAL_ABORT();
  }
  DEBUG_PRINTF("Expected %d, got it\n", token);
  tokenizer_next();
}
/*---------------------------------------------------------------------------*/
int varfactor(void)
{
  int r;
  DEBUG_PRINTF("varfactor: obtaining %d from variable %d\n", variables[tokenizer_variable_num()], tokenizer_variable_num());
  r = ubasic_get_variable(tokenizer_variable_num());
  accept(TOKENIZER_VARIABLE);
  return r;
}
/*---------------------------------------------------------------------------*/
int factor(void)
{
  int r = 0, i, *arr;
	T_USER_FUNC_DATAS temp;
	T_BASIC_USER_FUNC func;
	int type, data, idx;
	void *ptr;

  DEBUG_PRINTF("factor: token %d\n", tokenizer_token());
  switch(tokenizer_token()) {
  case TOKENIZER_NUMBER:
    r = tokenizer_num();
    DEBUG_PRINTF("factor: number %d\n", r);
    accept(TOKENIZER_NUMBER);
    break;
	case TOKENIZER_NAME:
		idx = tokenizer_userdata (&type, &data, &ptr);
		DEBUG_PRINTF("user keyword found : %d type %d\n", idx, data);
    switch (type)
		{
			case 1 : 
				accept (TOKENIZER_NAME);
				DEBUG_PRINTF("user function parameters\n");
				get_parameters (&temp, data, 0);
				func = ptr;
				r = func (temp);
				break;
			case 2 : 
				accept(TOKENIZER_NAME);
				DEBUG_PRINTF("user variable\n");
				r = *((int *) ptr);
				break;
			case 3 : 
				accept(TOKENIZER_NAME);
				DEBUG_PRINTF("user array\n");
				i = get_arrayindex ();
				arr = (int *) ptr;
				if ((i >= 0) && (i < data))
				{
					r = arr [i];
				}
				break;
		}
		break;
  case TOKENIZER_LEFTPAREN:
    accept(TOKENIZER_LEFTPAREN);
    r = expr();
    accept(TOKENIZER_RIGHTPAREN);
    break;
  default:
    r = varfactor();
    break;
  }
  return r;
}
/*---------------------------------------------------------------------------*/

#define BASIC_TOKEN_OPS

int term(void)
{
  int f1, f2;
  int op;

  f1 = factor();
  op = tokenizer_token();
  DEBUG_PRINTF("term: token %d\n", op);
	while (
		#define TOKEN_CHAR_TERM(x)	(op == x) ||
		#define TOKEN_CHAR_EXPR(x)
		#define TOKEN_CHAR_REL(x)
			#include BASIC_CONFIG_FILE
		#undef TOKEN_CHAR_TERM
		#undef TOKEN_CHAR_EXPR
		#undef TOKEN_CHAR_REL
		(0)) {

    tokenizer_next();
    f2 = factor();
    DEBUG_PRINTF("term: %d %d %d\n", f1, op, f2);
		f1 = execbasicexprfunc (op, f1, f2);
    op = tokenizer_token();
  }
  DEBUG_PRINTF("term: %d\n", f1);
  return f1;
}
/*---------------------------------------------------------------------------*/
int expr(void)
{
  int t1, t2;
  int op;
  
  t1 = term();
  op = tokenizer_token();
  DEBUG_PRINTF("expr: token %d\n", op);
	while (
		#define TOKEN_CHAR_TERM(x) 
		#define TOKEN_CHAR_EXPR(x)	(op == x) ||
		#define TOKEN_CHAR_REL(x)
			#include BASIC_CONFIG_FILE
		#undef TOKEN_CHAR_TERM
		#undef TOKEN_CHAR_EXPR
		#undef TOKEN_CHAR_REL
		(0)) {
    tokenizer_next();
    t2 = term();
    DEBUG_PRINTF("expr: %d %d %d\n", t1, op, t2);
		t1 = execbasicexprfunc (op, t1, t2);
    op = tokenizer_token();
  }
  DEBUG_PRINTF("expr: %d\n", t1);
  return t1;
}
/*---------------------------------------------------------------------------*/
int relation(void)
{
  int r1, r2;
  int op;
  
  r1 = expr();
  op = tokenizer_token();
  DEBUG_PRINTF("relation: token %d\n", op);
	while (
		#define TOKEN_CHAR_TERM(x) 
		#define TOKEN_CHAR_EXPR(x)
		#define TOKEN_CHAR_REL(x)		(op == x) ||
			#include BASIC_CONFIG_FILE
		#undef TOKEN_CHAR_TERM
		#undef TOKEN_CHAR_EXPR
		#undef TOKEN_CHAR_REL
		(0)) {
    tokenizer_next();
    r2 = expr();
    DEBUG_PRINTF("relation: %d %d %d\n", r1, op, r2);
		r1 = execbasicexprfunc (op, r1, r2);
    op = tokenizer_token();
  }
  return r1;
}


#undef BASIC_TOKEN_OPS

/*---------------------------------------------------------------------------*/
void jump_linenum(int linenum)
{
  tokenizer_init(program_ptr);
  while(tokenizer_num() != linenum) {
    do {
      do {
				tokenizer_next();
				if (tokenizer_token() == TOKENIZER_ERROR)
				{
					FATAL_ABORT();
				}
      } while(tokenizer_token() != TOKENIZER_CR &&
	      tokenizer_token() != TOKENIZER_ENDOFINPUT);
      if(tokenizer_token() == TOKENIZER_CR) {
				tokenizer_next();
      }
    } while(tokenizer_token() != TOKENIZER_NUMBER);
    DEBUG_PRINTF("jump_linenum: Found line %d\n", tokenizer_num());
  }
}
/*---------------------------------------------------------------------------*/
void goto_statement(void)
{
  accept(TOKENIZER_GOTO);
  jump_linenum(tokenizer_num());
}
/*---------------------------------------------------------------------------*/
void print_statement(void)
{
  accept(TOKENIZER_PRINT);
  do {
    DEBUG_PRINTF("Print loop\n");
    if(tokenizer_token() == TOKENIZER_STRING) {
      tokenizer_string(string, sizeof(string));
      printf("%s", string);
      tokenizer_next();
    } else if(tokenizer_token() == TOKENIZER_COMMA) {
      printf(" ");
      tokenizer_next();
    } else if(tokenizer_token() == TOKENIZER_SEMICOLON) {
      tokenizer_next();
    } else if(tokenizer_token() == TOKENIZER_VARIABLE ||
	      tokenizer_token() == TOKENIZER_NUMBER || 
				tokenizer_token() == TOKENIZER_NAME) {
      printf("%d", expr());
    } else {
      break;
    }
  } while(tokenizer_token() != TOKENIZER_CR &&
	  tokenizer_token() != TOKENIZER_ENDOFINPUT);
  printf("\n");
  DEBUG_PRINTF("End of print\n");
  tokenizer_next();
}
/*---------------------------------------------------------------------------*/
void if_statement(void)
{
  int r;
  
  accept(TOKENIZER_IF);

  r = relation();
  DEBUG_PRINTF("if_statement: relation %d\n", r);
  accept(TOKENIZER_THEN);
  if(r) {
    statement();
  } else {
    do {
      tokenizer_next();
    } while(tokenizer_token() != TOKENIZER_ELSE &&
	    tokenizer_token() != TOKENIZER_CR &&
	    tokenizer_token() != TOKENIZER_ENDOFINPUT);
    if(tokenizer_token() == TOKENIZER_ELSE) {
      tokenizer_next();
      statement();
    } else if(tokenizer_token() == TOKENIZER_CR) {
      tokenizer_next();
    }
  }
}
/*---------------------------------------------------------------------------*/
void let_statement(void)
{
  int var;

  var = tokenizer_variable_num();

  accept(TOKENIZER_VARIABLE);
  accept(TOKENIZER_EQ);
  ubasic_set_variable(var, expr());
  DEBUG_PRINTF("let_statement: assign %d to %d\n", variables[var], var);
  accept(TOKENIZER_CR);

}

/*---------------------------------------------------------------------------*/
void let_statement_2(void)
{
	accept (TOKENIZER_LET);
	let_statement ();
}
/*---------------------------------------------------------------------------*/
void gosub_statement(void)
{
  int linenum;
  accept(TOKENIZER_GOSUB);
  linenum = tokenizer_num();
  accept(TOKENIZER_NUMBER);
  accept(TOKENIZER_CR);
  if(gosub_stack_ptr < MAX_GOSUB_STACK_DEPTH) {
    gosub_stack[gosub_stack_ptr] = tokenizer_num();
    gosub_stack_ptr++;
    jump_linenum(linenum);
  } else {
    DEBUG_PRINTF("gosub_statement: gosub stack exhausted\n");
  }
}
/*---------------------------------------------------------------------------*/
void return_statement(void)
{
  accept(TOKENIZER_RETURN);
  if(gosub_stack_ptr > 0) {
    gosub_stack_ptr--;
    jump_linenum(gosub_stack[gosub_stack_ptr]);
  } else {
    DEBUG_PRINTF("return_statement: non-matching return\n");
  }
}
/*---------------------------------------------------------------------------*/
void next_statement(void)
{
  int var;
  
  accept(TOKENIZER_NEXT);
  var = tokenizer_variable_num();
  accept(TOKENIZER_VARIABLE);
  if(for_stack_ptr > 0 &&
     var == for_stack[for_stack_ptr - 1].for_variable) {
    ubasic_set_variable(var,
			ubasic_get_variable(var) + 1);
    if(ubasic_get_variable(var) <= for_stack[for_stack_ptr - 1].to) {
      jump_linenum(for_stack[for_stack_ptr - 1].line_after_for);
    } else {
      for_stack_ptr--;
      accept(TOKENIZER_CR);
    }
  } else {
    DEBUG_PRINTF("next_statement: non-matching next (expected %d, found %d)\n", for_stack[for_stack_ptr - 1].for_variable, var);
    accept(TOKENIZER_CR);
  }

}
/*---------------------------------------------------------------------------*/
void for_statement(void)
{
  int for_variable, to;
  
  accept(TOKENIZER_FOR);
  for_variable = tokenizer_variable_num();
  accept(TOKENIZER_VARIABLE);
  accept(TOKENIZER_EQ);
  ubasic_set_variable(for_variable, expr());
  accept(TOKENIZER_TO);
  to = expr();
  accept(TOKENIZER_CR);

  if(for_stack_ptr < MAX_FOR_STACK_DEPTH) {
    for_stack[for_stack_ptr].line_after_for = tokenizer_num();
    for_stack[for_stack_ptr].for_variable = for_variable;
    for_stack[for_stack_ptr].to = to;
    DEBUG_PRINTF("for_statement: new for, var %d to %d\n",
		 for_stack[for_stack_ptr].for_variable,
		 for_stack[for_stack_ptr].to);
		 
    for_stack_ptr++;
  } else {
    DEBUG_PRINTF("for_statement: for stack depth exceeded\n");
  }
}
/*---------------------------------------------------------------------------*/
void end_statement(void)
{
  accept(TOKENIZER_END);
  ended = 1;
}


/*---------------------------------------------------------------------------*/
void dummyfunc (void)
{
    DEBUG_PRINTF("ubasic.c: statement(): not implemented %d\n", token);
		FATAL_ABORT();
}
/*---------------------------------------------------------------------------*/
int dummyexprfunc (int a, int b)
{
	return 0;
}
/*---------------------------------------------------------------------------*/
void statement(void)
{
  int token;
	T_USER_FUNC_DATAS temp;
	T_BASIC_USER_FUNC func;
	int type, data, idx, *arr; 
	void *ptr;
  
  token = tokenizer_token();

	if (token == TOKENIZER_NAME)
	{
		accept (TOKENIZER_NAME);
		idx = tokenizer_userdata (&type, &data, &ptr);
		switch (type)
		{
			case 1 :
				get_parameters (&temp, data, 1);
				func = ptr;
				func (temp);
				break;
			case 3 :
				idx = get_arrayindex ();
				arr = ptr;
				accept (TOKENIZER_EQ);
				arr [idx] = expr();
				tokenizer_next();
				break;
			default : 
				FATAL_ABORT();
				break;
		}
	}
	else if (token == TOKENIZER_VARIABLE)
	{
		let_statement ();
	} 
	else 
	{
		execbasicfunc ();
	}

}
/*---------------------------------------------------------------------------*/
void line_statement(void)
{
  DEBUG_PRINTF("----------- Line number %d ---------\n", tokenizer_num());
  /*    current_linenum = tokenizer_num();*/
  accept(TOKENIZER_NUMBER);
  statement();
  return;
}
/*---------------------------------------------------------------------------*/
void ubasic_run(void)
{
  if(tokenizer_finished()) {
    DEBUG_PRINTF("uBASIC program finished\n");
    return;
  }

  line_statement();
}
/*---------------------------------------------------------------------------*/
int ubasic_finished(void)
{
  return ended || tokenizer_finished();
}
/*---------------------------------------------------------------------------*/
void ubasic_set_variable(int varnum, int value)
{
  if(varnum > 0 && varnum <= MAX_VARNUM) {
    variables[varnum - 1] = value;
  }
}
/*---------------------------------------------------------------------------*/
int ubasic_get_variable(int varnum)
{
  if(varnum > 0 && varnum <= MAX_VARNUM) {
    return variables[varnum - 1];
  }
  return 0;
}
/*---------------------------------------------------------------------------*/



int basic_add (int a, int b)
{
	return a+b;
}
int basic_sub (int a, int b)
{
	return a-b;
}
int basic_mul (int a, int b)
{
	return a*b;
}
int basic_div (int a, int b)
{
	return a/b;
}
int basic_mod (int a, int b)
{
	return a%b;
}
int basic_and (int a, int b)
{
	return a&b;
}
int basic_or (int a, int b)
{
	return a|b;
}
int basic_xor (int a, int b)
{
	return a^b;
}
int basic_rel_eq (int a, int b)
{
	return a==b;
}
int basic_rel_gt (int a, int b)
{
	return a>b;
}
int basic_rel_lt (int a, int b)
{
	return a<b;
}
int basic_rel_le (int a, int b)
{
	return a<=b;
}
int basic_rel_ge (int a, int b)
{
	return a>=b;
}
int basic_rel_ne (int a, int b)
{
	return a!=b;
}
int basic_shift_right (int a, int b)
{
	return a>>b;
}
int basic_shift_left (int a, int b)
{
	return a<<b;
}
